
function [desc , locs ]=features_description(f,im_features)
[M, N, D] = size(f);
f = uint8(f);
if D==3
    f = rgb2gray(f);
end
f = mat2gray(im2double(f));
g = gaussian_filter( 1.5 * 2 ); % 9 
zero_pad = ceil( length(g) / 2 ); % 5
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
fprintf( 'Computing gradient magnitude and orientation...\n' );
tic;
  diff_x = 0.5*(f(2:(end-1),3:(end))-f(2:(end-1),1:(end-2)));    %L(x+1,y)-L(x-1,y)
  diff_y = 0.5*(f(3:(end),2:(end-1))-f(1:(end-2),2:(end-1)));      %L(x,y+1)-L(x,y-1)
  mag = zeros(size(f));      
  mag(2:(end-1),2:(end-1)) = sqrt( diff_x .^ 2 + diff_y .^ 2 );
  mag_pad = zeros(size(mag)+2*zero_pad);
  mag_pad((zero_pad+1):(end-zero_pad),(zero_pad+1):(end-zero_pad)) = mag;      
  
  grad = zeros(size(f)); 
  grad(2:(end-1),2:(end-1)) = atan2( diff_y, diff_x );   
  grad(grad == pi) = -pi;
  grad_pad = zeros(size(grad)+2*zero_pad);
  grad_pad((zero_pad+1):(end-zero_pad),(zero_pad+1):(end-zero_pad)) = grad;
  clear mag grad
grad_time = toc;
fprintf( 'Gradient calculation time %.2f seconds.\n', grad_time );
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
fprintf( 'Assigining keypoint orientations...\n' );
num_bins = 36;
hist_step = 2*pi/num_bins; % 10"c
hist_orient = [-pi:hist_step:(pi-hist_step)];% -180:10:170
pos = [];        
orient = [];
scale = [];
tic;
    keypoint_count = 0;
  
    g = gaussian_filter( 1.5 * 2 ); 
    hf_sz = floor(length(g)/2);
    g = g'*g;      
    loc_pad = zeros(size(f)+2*zero_pad); 
    loc_pad((zero_pad+1):(end-zero_pad),(zero_pad+1):(end-zero_pad)) = im_features;
    [iy, ix]=find(loc_pad==1); 
    for k = 1:length(iy)
         x = ix(k);
         y = iy(k);
         wght = g.*mag_pad((y-hf_sz):(y+hf_sz),(x-hf_sz):(x+hf_sz));length(g)Xlength(g)
         grad_window = grad_pad((y-hf_sz):(y+hf_sz),(x-hf_sz):(x+hf_sz));
         orient_hist=zeros(length(hist_orient),1);% 36X1
         for bin=1:length(hist_orient)% 1 : 36
            diff = mod( grad_window - hist_orient(bin) + pi, 2*pi ) - pi; 
        
            orient_hist(bin)=orient_hist(bin)+sum(sum(wght.*max(1 - abs(diff)/hist_step,0))); 
         end
      
         peaks = orient_hist;     
         rot_right = [ peaks(end); peaks(1:end-1) ]; % 1-D array with 36 elements
         rot_left = [ peaks(2:end); peaks(1) ];  % 1-D array with 36 elements
         peaks(peaks < rot_right) = 0;
         peaks(peaks < rot_left) = 0;
   
         [max_peak_val, ipeak] = max(peaks); 
         peak_val = max_peak_val;
         while( peak_val > 0.8*max_peak_val )
           		            
            A = [];
            b = [];
            for j = -1:1
               A = [A; (hist_orient(ipeak)+hist_step*j).^2 (hist_orient(ipeak)+hist_step*j) 1];
               bin = mod( ipeak + j + num_bins - 1, num_bins ) + 1;  % num_bins=36;
               b = [b; orient_hist(bin)];
            end
            c = pinv(A)*b;
            max_orient = -c(2)/(2*c(1));  %c(1)x^2+c(2)x+c(3)
            max_wght = c(1)*max_orient^2 + c(2)*max_orient +c(3);
            while( max_orient < -pi )
               max_orient = max_orient + 2*pi;
            end
            while( max_orient >= pi )
               max_orient = max_orient - 2*pi;
            end            
     
            pos = [pos;(y-zero_pad) (x-zero_pad)];
            orient = [orient; max_orient];
            scale = [scale ; max_wght];
            keypoint_count = keypoint_count + 1;
            % Get the next peak
            peaks(ipeak) = 0;
            [peak_val, ipeak] = max(peaks);
         end         
    end
    fprintf( '(%d keypoints)\n', keypoint_count );       
    clear loc loc_pad 
orient_time = toc;
fprintf( 'Orientation assignment time %.2f seconds.\n', orient_time );
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
locs = [];
locs = [pos  scale orient];
showkeys(f, locs)
title('Final keypoints with scale and orientation');
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

orient_bin_spacing = pi/4;
orient_angles = [-pi:orient_bin_spacing:(pi-orient_bin_spacing)]; %  8-D  -pi:-pi/4:-pi*3/4
grid_spacing = 4;
[x_coords , y_coords] = meshgrid( [-2:grid_spacing:2] ); %-2,2 
feat_grid = [x_coords(:) y_coords(:)]'; 
[x_coords , y_coords] = meshgrid( [-(grid_spacing-0.5):(grid_spacing-0.5)] ); %8*8;-3.5:3.5
feat_samples = [x_coords(:) y_coords(:)]';  % 2*64
feat_window = grid_spacing; 
n = size(pos,1);
desc = zeros(n,32);
fprintf( 'Computing keypoint feature descriptors for %d keypoints', size(pos,1) );
tic;
    for k = 1:n
        if mod(k,100) == 0
            fprintf('%d',k);
        end
        x = pos(k,1);  
        y = pos(k,2);   
        
        M = [cos(orient(k)) -sin(orient(k)); sin(orient(k)) cos(orient(k))];
        feat_rot_grid = M*feat_grid + repmat([x; y],1,size(feat_grid,2));  
        feat_rot_samples = M*feat_samples + repmat([x; y],1,size(feat_samples,2)); 
        feat_desc = zeros(1,32);
        for s = 1:size(feat_rot_samples,2)  % 1:64   feat_rot_samples  2*256
             x_sample = feat_rot_samples(1,s); %-3.53.5
             y_sample = feat_rot_samples(2,s);
      
             [X,Y] = meshgrid( (x_sample-1):(x_sample+1), (y_sample-1):(y_sample+1) ); %-8.5-7.5-6.5-8.5-7.5-6.5
             G = interp2( f, X, Y, 'linear' );
             G(isnan(G)) = 0;  
             diff_x = 0.5*(G(2,3) - G(2,1));
             diff_y = 0.5*(G(3,2) - G(1,2));
             mag_sample = sqrt( diff_x^2 + diff_y^2 );
             grad_sample = atan2( diff_y, diff_x );
             if grad_sample == pi
                   grad_sample = -pi;
             end      
             x_wght = max(1 - (abs(feat_rot_grid(1,:) - x_sample)/grid_spacing), 0);
             y_wght = max(1 - (abs(feat_rot_grid(2,:) - y_sample)/grid_spacing), 0); % 1*4
             pos_wght = reshape(repmat(x_wght.*y_wght,8,1),1,32);% 1*32
             diff = mod( grad_sample - orient(k) - orient_angles + pi, 2*pi ) - pi;% 1*8 orient_angles[-pi:pi/4:pi*3/4] 
             orient_wght = max(1 - abs(diff)/orient_bin_spacing,0);
             orient_wght = repmat(orient_wght,1,4);      
             g = exp(-((x_sample-x)^2+(y_sample-y)^2)/(2*feat_window^2))/(2*pi*feat_window^2);% figure_window=4
             feat_desc = feat_desc + pos_wght.*orient_wght*g*mag_sample;  %1*32
        end
        feat_desc = feat_desc / norm(feat_desc);  
        feat_desc( feat_desc > 0.2 ) = 0.2;
        feat_desc = feat_desc / norm(feat_desc);
        desc(k,:) = feat_desc;   %n*128
    end
desc_time = toc;
fprintf('Description assignment time is %.2fs\n',desc_time);